/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <stdlib.h>
#include <getopt.h>
#include "spamdyke.h"
#include "dns.h"
#include "environment.h"
#include "command_line.h"

extern int opterr;

int append_string(char ***target_list, char *target_string, int strlen_target_string)
  {
  int return_value;
  int i;
  char *tmp_char;
  char **tmp_array;

  return_value = 0;

  if ((*target_list) != NULL)
    for (i = 0; (*target_list)[i] != NULL; i++);
  else
    i = 0;

  tmp_char = NULL;
  if ((strlen_target_string == -1) ||
      ((tmp_char = malloc(strlen_target_string + 1)) != NULL))
    if ((tmp_array = realloc((*target_list), sizeof(char *) * (i + 2))) != NULL)
      {
      if (strlen_target_string >= 0)
        {
        memcpy(tmp_char, target_string, strlen_target_string);
        tmp_char[strlen_target_string] = '\0';
        tmp_array[i] = tmp_char;
        }
      else
        tmp_array[i] = target_string;

      tmp_array[i + 1] = NULL;
      *target_list = tmp_array;

      return_value = 1;
      }
    else
      {
      syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (i + 2)));
      free(tmp_char);
      }
  else
    syslog(LOG_ERR, SYSLOG_ERROR_REALLOC, (unsigned long)(strlen_target_string + 1));

  return(return_value);
  }

int process_command_line(struct filter_settings *current_settings, int argc, char *argv[], char *envp[])
  {
  int return_value;
  int i;
  int j;
  int tmp_int;
  long tmp_long;
  int opt;
  struct option long_options[] = {
    { "reject-ip-in-cc-rdns",           no_argument,            NULL,   'c' },
    { "reject-empty-rdns",              no_argument,            NULL,   'r' },
    { "reject-unresolvable-rdns",       no_argument,            NULL,   'R' },
    { "reject-missing-sender-mx",       no_argument,            NULL,   256 },

    { "use-syslog",                     no_argument,            NULL,   'l' },
    { "full-log-dir",                   required_argument,      NULL,   'L' },
    { "policy-url",                     required_argument,      NULL,   'u' },
    { "connection-timeout-secs",        required_argument,      NULL,   't' },
    { "idle-timeout-secs",              required_argument,      NULL,   'T' },

    { "access-file",                    required_argument,      NULL,   263 },
    { "local-domains-file",             required_argument,      NULL,   'd' },
    { "max-recipients",                 required_argument,      NULL,   'a' },

    { "graylist-dir",                   required_argument,      NULL,   'g' },
    { "no-graylist-dir",                required_argument,      NULL,   'G' },
    { "graylist-min-secs",              required_argument,      NULL,   'm' },
    { "graylist-max-secs",              required_argument,      NULL,   'M' },
    { "never-graylist-ip-file",         required_argument,      NULL,   257 },
    { "never-graylist-rdns-file",       required_argument,      NULL,   258 },
    { "always-graylist-ip-file",        required_argument,      NULL,   259 },
    { "always-graylist-rdns-file",      required_argument,      NULL,   260 },
    { "smtp-auth-command",              required_argument,      NULL,   261 },
    { "smtp-auth-command-encryption",   required_argument,      NULL,   264 },

    { "greeting-delay-secs",            required_argument,      NULL,   'e' },

    { "check-dnsrbl",                   required_argument,      NULL,   'x' },

    { "ip-in-rdns-keyword-file",        required_argument,      NULL,   'k' },
    { "rdns-blacklist-dir",             required_argument,      NULL,   'b' },
    { "ip-blacklist-file",              required_argument,      NULL,   'B' },
    { "rdns-whitelist-file",            required_argument,      NULL,   'w' },
    { "ip-whitelist-file",              required_argument,      NULL,   'W' },
    { "sender-blacklist-file",          required_argument,      NULL,   's' },
    { "recipient-blacklist-file",       required_argument,      NULL,   'S' },

    { "hostname",                       required_argument,      NULL,   262 },

    { 0, 0, 0, 0 }
    };

  return_value = FILTER_DECISION_UNDECIDED;
  opterr = 0;

  current_settings->server_ip = find_environment_variable(envp, ENVIRONMENT_REMOTE_IP, STRLEN(ENVIRONMENT_REMOTE_IP));
  current_settings->strlen_server_ip = (current_settings->server_ip != NULL) ? strlen(current_settings->server_ip) : 0;
  dns_ptr(current_settings->server_ip, current_settings);

  if (find_environment_variable(envp, ENVIRONMENT_TLS, STRLEN(ENVIRONMENT_TLS)))
    return_value = FILTER_DECISION_DO_NOT_FILTER;
  else
    while ((return_value != FILTER_DECISION_ERROR) &&
           ((opt = getopt_long(argc, argv, "+clrRa:b:B:d:e:g:G:k:L:m:M:s:S:t:T:w:W:u:x:", long_options, NULL)) != -1))
      {
      switch (opt)
        {
        case 'a':
          if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
              (tmp_int >= 0))
            {
            current_settings->max_rcpt_to = tmp_int;
            return_value = FILTER_DECISION_DO_FILTER;
            }

          break;
        case 'b':
          return_value = append_string(&current_settings->blacklist_rdns_directory_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'B':
          return_value = append_string(&current_settings->blacklist_ip_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'c':
          current_settings->check_ip_in_rdns_cc = 1;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'd':
          if (!append_string(&current_settings->local_domains_path, optarg, -1))
            return_value = FILTER_DECISION_ERROR;

          break;
        case 'e':
          if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
              (tmp_int >= 0))
            {
            current_settings->check_earlytalker = tmp_int;
            return_value = FILTER_DECISION_DO_FILTER;
            }

          break;
        case 'g':
          current_settings->graylist_type = GRAYLIST_FILTER_ALL;
          current_settings->graylist_dir = optarg;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'G':
          current_settings->graylist_type = GRAYLIST_FILTER_NONE;
          current_settings->graylist_dir = optarg;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 257:
          return_value = append_string(&current_settings->graylist_never_ip_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 258:
          return_value = append_string(&current_settings->graylist_never_rdns_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 259:
          return_value = append_string(&current_settings->graylist_always_ip_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 260:
          return_value = append_string(&current_settings->graylist_always_rdns_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'k':
          return_value = append_string(&current_settings->blacklist_rdns_keyword_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'l':
          current_settings->use_syslog = 1;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'L':
          current_settings->log_dir = optarg;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'm':
          if ((sscanf(optarg, "%ld", &tmp_long) == 1) &&
              (tmp_long >= 0))
            current_settings->graylist_min_secs = tmp_long;

          break;
        case 'M':
          if ((sscanf(optarg, "%ld", &tmp_long) == 1) &&
              (tmp_long >= 0))
            current_settings->graylist_max_secs = tmp_long;

          break;
        case 'r':
          current_settings->check_rdns_exist = 1;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'R':
          current_settings->check_rdns_resolve = 1;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 's':
          current_settings->blacklist_sender_path = optarg;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 'S':
          current_settings->blacklist_recipient_path = optarg;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 't':
          if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
              (tmp_int >= 0))
            {
            current_settings->timeout_connection = tmp_int;
            return_value = FILTER_DECISION_DO_FILTER;
            }

          break;
        case 'T':
          if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
              (tmp_int >= 0))
            {
            current_settings->timeout_command = tmp_int;
            return_value = FILTER_DECISION_DO_FILTER;
            }

          break;
        case 'w':
          return_value = append_string(&current_settings->whitelist_rdns_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'W':
          return_value = append_string(&current_settings->whitelist_ip_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'x':
          return_value = append_string(&current_settings->dnsrbl_fqdn, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 'u':
          if ((current_settings->policy_location = optarg) != NULL)
            current_settings->strlen_policy_location = MINVAL(MAX_POLICY, strlen(optarg));

          break;
        case 256:
          current_settings->check_sender_mx = 1;
          return_value = FILTER_DECISION_DO_FILTER;

          break;
        case 261:
          return_value = append_string(&current_settings->smtp_auth_command, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;
          current_settings->smtp_auth_encryption = 0;

          break;
        case 262:
          current_settings->local_server_name = optarg;

          break;
        case 263:
          return_value = append_string(&current_settings->access_list_path, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;

          break;
        case 264:
          return_value = append_string(&current_settings->smtp_auth_command, optarg, -1) ? FILTER_DECISION_DO_FILTER : FILTER_DECISION_ERROR;
          current_settings->smtp_auth_encryption = 1;

          break;
        default:
          return_value = FILTER_DECISION_ERROR;
          break;
        }
      }

  if (current_settings->smtp_auth_command != NULL)
    for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
      for (j = 0; current_settings->smtp_auth_command[i][j] != '\0'; j++)
        if (current_settings->smtp_auth_command[i][j] == SMTP_AUTH_SPACER)
          current_settings->smtp_auth_command[i][j] = ' ';

  argv += optind;
  if (*argv != NULL)
    current_settings->command_index = optind;

  return(return_value);
  }
